# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 6.3.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGroupBox, QLabel,
    QLineEdit, QMainWindow, QMenuBar, QPushButton,
    QSizePolicy, QStatusBar, QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(590, 692)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.label = QLabel(self.centralwidget)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(30, 210, 121, 16))
        self.geschwindigkeit = QLineEdit(self.centralwidget)
        self.geschwindigkeit.setObjectName(u"geschwindigkeit")
        self.geschwindigkeit.setGeometry(QRect(150, 210, 61, 21))
        self.geschwindigkeit.setReadOnly(True)
        self.label_2 = QLabel(self.centralwidget)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(30, 240, 101, 16))
        self.druckhll = QLineEdit(self.centralwidget)
        self.druckhll.setObjectName(u"druckhll")
        self.druckhll.setGeometry(QRect(150, 240, 201, 21))
        self.druckhll.setReadOnly(True)
        self.groupBox = QGroupBox(self.centralwidget)
        self.groupBox.setObjectName(u"groupBox")
        self.groupBox.setGeometry(QRect(20, 300, 171, 121))
        self.lm_u = QLabel(self.groupBox)
        self.lm_u.setObjectName(u"lm_u")
        self.lm_u.setGeometry(QRect(10, 10, 51, 51))
        self.lm_u.setAutoFillBackground(False)
        self.lm_u.setStyleSheet(u"background-color: blue;")
        self.lm_u.setAlignment(Qt.AlignCenter)
        self.lm_m = QLabel(self.groupBox)
        self.lm_m.setObjectName(u"lm_m")
        self.lm_m.setGeometry(QRect(60, 10, 51, 51))
        self.lm_m.setAutoFillBackground(False)
        self.lm_m.setStyleSheet(u"background-color: blue;")
        self.lm_m.setAlignment(Qt.AlignCenter)
        self.lm_o = QLabel(self.groupBox)
        self.lm_o.setObjectName(u"lm_o")
        self.lm_o.setGeometry(QRect(110, 10, 51, 51))
        self.lm_o.setAutoFillBackground(False)
        self.lm_o.setStyleSheet(u"background-color: blue;")
        self.lm_o.setAlignment(Qt.AlignCenter)
        self.lm_40 = QLabel(self.groupBox)
        self.lm_40.setObjectName(u"lm_40")
        self.lm_40.setGeometry(QRect(10, 60, 51, 51))
        self.lm_40.setAutoFillBackground(False)
        self.lm_40.setStyleSheet(u"background-color: black")
        self.lm_40.setAlignment(Qt.AlignCenter)
        self.lm_500 = QLabel(self.groupBox)
        self.lm_500.setObjectName(u"lm_500")
        self.lm_500.setGeometry(QRect(60, 60, 51, 51))
        self.lm_500.setAutoFillBackground(False)
        self.lm_500.setStyleSheet(u"background-color: red;")
        self.lm_500.setAlignment(Qt.AlignCenter)
        self.lm_1000 = QLabel(self.groupBox)
        self.lm_1000.setObjectName(u"lm_1000")
        self.lm_1000.setGeometry(QRect(110, 60, 51, 51))
        self.lm_1000.setAutoFillBackground(False)
        self.lm_1000.setStyleSheet(u"background-color: yellow;")
        self.lm_1000.setAlignment(Qt.AlignCenter)
        self.lm_sifa = QLabel(self.centralwidget)
        self.lm_sifa.setObjectName(u"lm_sifa")
        self.lm_sifa.setGeometry(QRect(230, 310, 51, 51))
        self.lm_sifa.setAutoFillBackground(False)
        self.lm_sifa.setStyleSheet(u"background-color: gray;")
        self.lm_sifa.setAlignment(Qt.AlignCenter)
        self.label_10 = QLabel(self.centralwidget)
        self.label_10.setObjectName(u"label_10")
        self.label_10.setGeometry(QRect(30, 440, 91, 16))
        self.indusistatus = QLineEdit(self.centralwidget)
        self.indusistatus.setObjectName(u"indusistatus")
        self.indusistatus.setGeometry(QRect(150, 440, 201, 21))
        self.indusistatus.setAlignment(Qt.AlignCenter)
        self.indusistatus.setReadOnly(True)
        self.label_11 = QLabel(self.centralwidget)
        self.label_11.setObjectName(u"label_11")
        self.label_11.setGeometry(QRect(30, 510, 111, 16))
        self.sifastatus = QLineEdit(self.centralwidget)
        self.sifastatus.setObjectName(u"sifastatus")
        self.sifastatus.setGeometry(QRect(150, 510, 201, 21))
        self.sifastatus.setAlignment(Qt.AlignCenter)
        self.sifastatus.setReadOnly(True)
        self.label_12 = QLabel(self.centralwidget)
        self.label_12.setObjectName(u"label_12")
        self.label_12.setGeometry(QRect(30, 180, 121, 16))
        self.uhrzeit = QLineEdit(self.centralwidget)
        self.uhrzeit.setObjectName(u"uhrzeit")
        self.uhrzeit.setGeometry(QRect(150, 180, 201, 21))
        self.uhrzeit.setReadOnly(True)
        self.label_13 = QLabel(self.centralwidget)
        self.label_13.setObjectName(u"label_13")
        self.label_13.setGeometry(QRect(30, 540, 111, 16))
        self.nbuestatus = QLineEdit(self.centralwidget)
        self.nbuestatus.setObjectName(u"nbuestatus")
        self.nbuestatus.setGeometry(QRect(150, 540, 201, 21))
        self.nbuestatus.setAlignment(Qt.AlignCenter)
        self.nbuestatus.setReadOnly(True)
        self.sifabauart = QLineEdit(self.centralwidget)
        self.sifabauart.setObjectName(u"sifabauart")
        self.sifabauart.setGeometry(QRect(360, 510, 211, 21))
        self.sifabauart.setAlignment(Qt.AlignCenter)
        self.sifabauart.setReadOnly(True)
        self.indusibauart = QLineEdit(self.centralwidget)
        self.indusibauart.setObjectName(u"indusibauart")
        self.indusibauart.setGeometry(QRect(360, 440, 211, 21))
        self.indusibauart.setAlignment(Qt.AlignCenter)
        self.indusibauart.setReadOnly(True)
        self.nbuebauart = QLineEdit(self.centralwidget)
        self.nbuebauart.setObjectName(u"nbuebauart")
        self.nbuebauart.setGeometry(QRect(360, 540, 211, 21))
        self.nbuebauart.setAlignment(Qt.AlignCenter)
        self.nbuebauart.setReadOnly(True)
        self.indusizwangsbremsung = QLineEdit(self.centralwidget)
        self.indusizwangsbremsung.setObjectName(u"indusizwangsbremsung")
        self.indusizwangsbremsung.setGeometry(QRect(150, 470, 201, 21))
        self.indusizwangsbremsung.setAlignment(Qt.AlignCenter)
        self.indusizwangsbremsung.setReadOnly(True)
        self.label_14 = QLabel(self.centralwidget)
        self.label_14.setObjectName(u"label_14")
        self.label_14.setGeometry(QRect(30, 470, 111, 16))
        self.indusizusatzinfo = QLineEdit(self.centralwidget)
        self.indusizusatzinfo.setObjectName(u"indusizusatzinfo")
        self.indusizusatzinfo.setGeometry(QRect(360, 470, 211, 21))
        self.indusizusatzinfo.setAlignment(Qt.AlignCenter)
        self.indusizusatzinfo.setReadOnly(True)
        self.sollgeschwindigkeit = QLineEdit(self.centralwidget)
        self.sollgeschwindigkeit.setObjectName(u"sollgeschwindigkeit")
        self.sollgeschwindigkeit.setGeometry(QRect(270, 210, 81, 21))
        self.sollgeschwindigkeit.setReadOnly(True)
        self.label_15 = QLabel(self.centralwidget)
        self.label_15.setObjectName(u"label_15")
        self.label_15.setGeometry(QRect(30, 590, 111, 16))
        self.tuerenlinks = QLineEdit(self.centralwidget)
        self.tuerenlinks.setObjectName(u"tuerenlinks")
        self.tuerenlinks.setGeometry(QRect(150, 590, 201, 21))
        self.tuerenlinks.setAlignment(Qt.AlignCenter)
        self.tuerenlinks.setReadOnly(True)
        self.tuerenrechts = QLineEdit(self.centralwidget)
        self.tuerenrechts.setObjectName(u"tuerenrechts")
        self.tuerenrechts.setGeometry(QRect(360, 590, 211, 21))
        self.tuerenrechts.setAlignment(Qt.AlignCenter)
        self.tuerenrechts.setReadOnly(True)
        self.label_16 = QLabel(self.centralwidget)
        self.label_16.setObjectName(u"label_16")
        self.label_16.setGeometry(QRect(30, 20, 121, 16))
        self.zusi_ip = QLineEdit(self.centralwidget)
        self.zusi_ip.setObjectName(u"zusi_ip")
        self.zusi_ip.setGeometry(QRect(150, 20, 151, 21))
        self.connectButton = QPushButton(self.centralwidget)
        self.connectButton.setObjectName(u"connectButton")
        self.connectButton.setGeometry(QRect(330, 10, 100, 41))
        self.zugnummer = QLineEdit(self.centralwidget)
        self.zugnummer.setObjectName(u"zugnummer")
        self.zugnummer.setGeometry(QRect(150, 100, 201, 21))
        self.zugnummer.setReadOnly(True)
        self.label_17 = QLabel(self.centralwidget)
        self.label_17.setObjectName(u"label_17")
        self.label_17.setGeometry(QRect(30, 100, 121, 16))
        self.fahrplan = QLineEdit(self.centralwidget)
        self.fahrplan.setObjectName(u"fahrplan")
        self.fahrplan.setGeometry(QRect(30, 70, 531, 21))
        self.fahrplan.setReadOnly(True)
        self.label_18 = QLabel(self.centralwidget)
        self.label_18.setObjectName(u"label_18")
        self.label_18.setGeometry(QRect(30, 50, 121, 16))
        self.ladezustand = QLineEdit(self.centralwidget)
        self.ladezustand.setObjectName(u"ladezustand")
        self.ladezustand.setGeometry(QRect(150, 130, 201, 21))
        self.ladezustand.setReadOnly(True)
        self.label_19 = QLabel(self.centralwidget)
        self.label_19.setObjectName(u"label_19")
        self.label_19.setGeometry(QRect(30, 130, 121, 16))
        self.bremsdruck = QLineEdit(self.centralwidget)
        self.bremsdruck.setObjectName(u"bremsdruck")
        self.bremsdruck.setGeometry(QRect(150, 270, 201, 21))
        self.bremsdruck.setReadOnly(True)
        self.label_3 = QLabel(self.centralwidget)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setGeometry(QRect(30, 270, 101, 16))
        self.label_4 = QLabel(self.centralwidget)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setGeometry(QRect(230, 210, 51, 16))
        self.sollgeschwindigkeit_check = QCheckBox(self.centralwidget)
        self.sollgeschwindigkeit_check.setObjectName(u"sollgeschwindigkeit_check")
        self.sollgeschwindigkeit_check.setGeometry(QRect(360, 210, 85, 20))
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 590, 33))
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"ZusiData", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Geschwindigkeit:", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Druck HLL:", None))
        self.groupBox.setTitle("")
        self.lm_u.setText(QCoreApplication.translate("MainWindow", u"55", None))
        self.lm_m.setText(QCoreApplication.translate("MainWindow", u"70", None))
        self.lm_o.setText(QCoreApplication.translate("MainWindow", u"85", None))
        self.lm_40.setText(QCoreApplication.translate("MainWindow", u"Befehl\n"
"40", None))
        self.lm_500.setText(QCoreApplication.translate("MainWindow", u"500Hz", None))
        self.lm_1000.setText(QCoreApplication.translate("MainWindow", u"1000Hz", None))
        self.lm_sifa.setText(QCoreApplication.translate("MainWindow", u"Sifa", None))
        self.label_10.setText(QCoreApplication.translate("MainWindow", u"Indusi Status:", None))
        self.label_11.setText(QCoreApplication.translate("MainWindow", u"Sifa Status:", None))
        self.label_12.setText(QCoreApplication.translate("MainWindow", u"Uhrzeit:", None))
        self.label_13.setText(QCoreApplication.translate("MainWindow", u"Notbremsstatus:", None))
        self.label_14.setText(QCoreApplication.translate("MainWindow", u"Zwangsbremsung:", None))
        self.label_15.setText(QCoreApplication.translate("MainWindow", u"T\u00fcrstatus:", None))
        self.label_16.setText(QCoreApplication.translate("MainWindow", u"Zusi IP", None))
        self.zusi_ip.setText(QCoreApplication.translate("MainWindow", u"127.0.0.1", None))
        self.connectButton.setText(QCoreApplication.translate("MainWindow", u"Connect", None))
        self.label_17.setText(QCoreApplication.translate("MainWindow", u"Zugnummer:", None))
        self.label_18.setText(QCoreApplication.translate("MainWindow", u"Fahrplan:", None))
        self.label_19.setText(QCoreApplication.translate("MainWindow", u"Ladezustand:", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Bremsdruck:", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Max:", None))
        self.sollgeschwindigkeit_check.setText(QCoreApplication.translate("MainWindow", u"zeigen", None))
    # retranslateUi




from enum import Enum
import struct
import logging
import asyncio

from pyzusi3.exceptions import DecodeValueError, EncodingValueError, MissingBytesDecodeError, MissingContentTypeError

class ContentType(Enum):
    BYTE = 0
    SHORTINT = 1
    WORD = 2
    SMALLINT = 3
    INTEGER = 4
    CARDINAL = 5
    INTEGER64BIT = 6
    SINGLE = 7
    DOUBLE = 8
    STRING = 9
    FILE = 10
    RAW = 11 # same as file but used to indicate non-decoded content


class BasicNode:
    def __init__(self, id=None, content=None, contenttype=None, children=None, parent_node=None, nodeasbool=False) -> None:
        self.id = id
        self.content = content
        self.contenttype = contenttype
        if children is not None:
            self.children = children
        else:
            self.children = []
        self.parent_node = parent_node
        self.nodeasbool = nodeasbool

    def __lt__(self, other):
        return self.id < other.id

    def __eq__(self, other):
        return other is not None and self.id == other.id and self.content == other.content and self.contenttype == other.contenttype and self.children == other.children and self.nodeasbool == other.nodeasbool

    def __repr__(self):
        return "<%s id=%s content=%s contenttype=%s parent_node=%s children_len=%s, nodeasbool=%s>" % (self.__class__.__name__, self.id, self.content, self.contenttype, self.parent_node, len(self.children), self.nodeasbool)

    def _encodecontent(self):
        result = b''
        if self.contenttype is None:
            raise MissingContentTypeError("No contenttype has been given to encode to")
        if self.contenttype == ContentType.BYTE:
            try:
                if not 0 <= self.content <= 255:
                    raise EncodingValueError("Content %s exceeds limits of 0-255 for a byte" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range 0-255" % str(self.content))
            result += self.content.to_bytes(1, byteorder='little')
        elif self.contenttype == ContentType.SHORTINT:
            try:
                if not -128 <= self.content <= 127:
                    raise EncodingValueError("Content %s exceeds limits of -128 to 127 for a short int" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -128 to 127" % str(self.content))
            result += self.content.to_bytes(1, byteorder='little')
        elif self.contenttype == ContentType.WORD:
            try:
                if not 0 <= self.content <= 65535:
                    raise EncodingValueError("Content %s exceeds limits of 0-65535 for a word" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range 0-65535" % str(self.content))
            result += self.content.to_bytes(2, byteorder='little')
        elif self.contenttype == ContentType.SMALLINT:
            try:
                if not -32768 <= self.content <= 32767:
                    raise EncodingValueError("Content %s exceeds limits of -32768 to 32767 for a small int" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -32768 to 32767" % str(self.content))
            result += self.content.to_bytes(2, byteorder='little')
        elif self.contenttype == ContentType.INTEGER:
            try:
                if not -2147483648 <= self.content <= 2147483647:
                    raise EncodingValueError("Content %s exceeds limits of -2147483648 to 2147483647 for an int" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -2147483648 to 2147483647" % str(self.content))
            result += self.content.to_bytes(4, byteorder='little')
        elif self.contenttype == ContentType.CARDINAL:
            try:
                if not 0 <= self.content <= 4294967295:
                    raise EncodingValueError("Content %s exceeds limits of 0 to 4294967295 for a cardinal" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range 0 to 4294967295" % str(self.content))
            result += self.content.to_bytes(4, byteorder='little')
        elif self.contenttype == ContentType.INTEGER64BIT:
            try:
                if not -9223372036854775808 <= self.content <= 9223372036854775807:
                    raise EncodingValueError("Content %s exceeds limits of -9223372036854775808 to 9223372036854775807 for an int64" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -9223372036854775808 to 9223372036854775807" % str(self.content))
            result += self.content.to_bytes(8, byteorder='little')
        elif self.contenttype == ContentType.SINGLE:
            try:
                if not -3.4E38 <= self.content <= 3.4E38:
                    raise EncodingValueError("Content %s exceeds limits of -3.4E38 to 3.4E38 for a single" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -3.4E38 to 3.4E38" % str(self.content))
            try:
                result += struct.pack("<f", self.content)
            except OverflowError as e:
                raise EncodingValueError("Content %s cannot be encoded: %s" % str(e))
        elif self.contenttype == ContentType.DOUBLE:
            try:
                if not -1.7E308 <= self.content <= 1.7E308:
                    raise EncodingValueError("Content %s exceeds limits of -1.7E308 to 1.7E308 for a double" % str(self.content))
            except TypeError:
                raise EncodingValueError("Content %s cannot be compared for range -1.7E308 to 1.7E308" % str(self.content))
            try:
                result += struct.pack("<d", self.content)
            except OverflowError as e:
                raise EncodingValueError("Content %s cannot be encoded: %s" % str(e))
        elif self.contenttype == ContentType.STRING:
            try:
                result += self.content.encode("latin1")
            except UnicodeEncodeError:
                raise EncodingValueError("Content %s cannot be encoded: %s" % str(e))
        elif self.contenttype == ContentType.FILE or self.contenttype == ContentType.RAW:
            if not isinstance(self.content, bytes):
                raise EncodingValueError("Content is not in bytes format")
            result += self.content
        else:
            raise MissingContentTypeError("Content of type %s is unknown in the encoder. Programming bug?" % (self.contenttype))

        return result

    def encode(self):
        result = b''

        if self.children or (not self.children and self.content is None) or self.nodeasbool:
            result = (0).to_bytes(4, byteorder='little') # Node start

        if self.content is not None:
            bytecontent = self._encodecontent()
            bytecontentlength = 2 + len(bytecontent) # add 2 for the content id
            result += bytecontentlength.to_bytes(4, byteorder='little')
        if self.id is not None:
            result += self.id.to_bytes(2, byteorder='little')
        if self.content is not None:
            result += bytecontent

        if self.children or (not self.children and self.content is None) or self.nodeasbool:
            for child in self.children:
                result += child.encode()
            result += (0xffffffff).to_bytes(4, byteorder='little') # Node end
        
        return result


class DecoderState(Enum):
    RESET = 0
    CONTRENTLENGTH = 1
    NODEID = 2
    NODEIDFORCONTENT = 3
    CONTENT = 4


class StreamDecoder:
    def __init__(self) -> None:
        self.log = logging.getLogger("pyzusi3.nodes.StreamDecoder")
        self.logstream = logging.getLogger("pyzusi3.nodes.StreamDecoder.streamdata")

    def reset(self):
        self.log.info("Resetting state")
        self.state = DecoderState.RESET
        self.root_node = None
        self.current_node = None
        self.content_length = None
        self.level = 1

    def decode(self, bytecontent):
        if not isinstance(bytecontent, bytes):
            raise ValueError("Need bytes to decode, not %s" % type(bytecontent))
        self.bytecontent = iter(bytecontent)
        self.log.info("Start decoding")
        while True:
            self.reset()
            self._decode_loop()
            self.log.debug("Decoding result:")
            self.log.debug(repr(self.root_node))
            if self.root_node is None:
                break
            yield self.root_node

    def _next_content_length(self):
        if self.state == DecoderState.RESET:
            length = 4
        elif self.state == DecoderState.NODEID or self.state == DecoderState.NODEIDFORCONTENT:
            length = 2
        elif self.state == DecoderState.CONTRENTLENGTH:
            length = 4
        elif self.state == DecoderState.CONTENT:
            length = self.content_length
        else:
            raise NotImplementedError("State %s not implemented in length check" % self.state)

        return length

    def _get_bytes(self):
        data = []
        for i in range(self._next_content_length()):
            try:
                data.append(next(self.bytecontent))
            except StopIteration:
                raise MissingBytesDecodeError("Not enough bytes to get from datastream")
        return bytes(data)

    def _decode_loop(self):
        while self.level > 0:
            try:
                incoming_bytes = self._get_bytes()
            except MissingBytesDecodeError as e:
                if self.level == 1:
                    break
                else:
                    raise MissingBytesDecodeError(str(e))
            self._decode_single_pass(incoming_bytes)
            if self.state == DecoderState.CONTRENTLENGTH and self.level == 1:
                # finished decoding
                break
        if self.current_node != self.root_node and self.current_node != None:
            raise DecodeValueError("Not all nodes have been closed, data incomplete")

    def _decode_single_pass(self, incoming_bytes):
        self.log.debug("Current state: %s" % self.state)
        if self.state == DecoderState.RESET:
            if incoming_bytes != (0).to_bytes(4, byteorder='little'):
                raise DecodeValueError("Expected 4 empty bytes for root node start but got %s" % incoming_bytes)
            self.root_node = self.current_node = BasicNode()
            self.log.debug("Created new node")
            self.state = DecoderState.NODEID
            self.level += 1
            self.log.debug("New level: %s" % self.level)

        elif self.state == DecoderState.NODEID or self.state == DecoderState.NODEIDFORCONTENT:
            self.current_node.id = struct.unpack("<H", incoming_bytes)[0]
            self.log.debug("Set node id to %s" % self.current_node.id)
            if self.state == DecoderState.NODEIDFORCONTENT:
                self.state = DecoderState.CONTENT
            else:
                self.state = DecoderState.CONTRENTLENGTH
        elif self.state == DecoderState.CONTRENTLENGTH:
            if incoming_bytes == (0xffffffff).to_bytes(4, byteorder='little'):
                # marks end of nodetree
                self.log.debug("Finished current node")
                if not self.current_node.children and self.current_node.content is None:
                    self.current_node.nodeasbool = True
                self.current_node = self.current_node.parent_node
                self.state = DecoderState.CONTRENTLENGTH
                self.level -= 1
                self.log.debug("New level: %s" % self.level)
                return
            self.log.debug("Creating new child node")
            new_node = BasicNode(parent_node=self.current_node)
            self.current_node.children.append(new_node)
            self.current_node = new_node
            content_length = struct.unpack("<I", incoming_bytes)[0]
            if content_length == 0:
                # just subnodes, possibly with children themselves
                self.state = DecoderState.NODEID
                self.log.debug("No content, just setting id afterwards")
                self.level += 1
                self.log.debug("New level: %s" % self.level)
            else:
                # real content follows after id
                self.content_length = struct.unpack("<I", incoming_bytes)[0] - 2
                self.log.debug("Setting id and adding content with length %s to node" % self.content_length)
                self.state = DecoderState.NODEIDFORCONTENT
        elif self.state == DecoderState.CONTENT:
            self.current_node.content = incoming_bytes
            self.current_node.contenttype = ContentType.RAW
            self.log.debug("Got content %s" % self.current_node.content)
            self.current_node = self.current_node.parent_node
            self.state = DecoderState.CONTRENTLENGTH

class AsyncStreamDecoder(StreamDecoder):

    async def _get_bytes(self):
        length = self._next_content_length()
        self.logstream.debug("Next content length: %s" % length)
        data = b''
        while len(data) != length:
            new_data = b''
            while True:
                new_data = await self.bytecontent.readexactly(length)
                break
            if new_data == b'':
                # eof
                break
            data += new_data
        self.logstream.debug("Got data: %s" % data)
        return data

    async def decode(self, stream):
        if not isinstance(stream, asyncio.StreamReader):
            raise ValueError("Need stream to decode, not %s" % type(stream))
        self.bytecontent = stream
        self.log.info("Start decoding")
        while True:
            self.reset()
            try:
                await self._decode_loop()
            except asyncio.exceptions.IncompleteReadError as e:
                if self.level == 1:
                    break
                else:
                    raise asyncio.exceptions.IncompleteReadError(str(e))                
            self.log.debug("Decoding result:")
            self.log.debug(repr(self.root_node))
            if self.root_node is None:
                break
            yield self.root_node

    async def _decode_loop(self):
        while self.level > 0:
            incoming_bytes = await self._get_bytes()
            self.log.debug("Next decode pass")
            self._decode_single_pass(incoming_bytes)
            if self.state == DecoderState.CONTRENTLENGTH and self.level == 1:
                # finished decoding
                break
        if self.current_node != self.root_node and self.current_node != None:
            raise DecodeValueError("Not all nodes have been closed, data incomplete")



import functools
import logging
import sys

from PySide6.QtWidgets import QApplication
from mainwindow import MainWindow

#logging.basicConfig(level=logging.INFO)

# Important:
# You need to run the following command to (re)generate the ui_form.py file
#     pyside6-uic form.ui -o ui_form.py

def main():
    app = QApplication(sys.argv)

    mainWindow = MainWindow()
    mainWindow.show()

    app.exec()
    sys.exit()

if __name__ == "__main__":
    main()


import asyncio
import unittest

from pyzusi3.messagecoders import MessageDecoder, encode_obj
from pyzusi3 import messages
from pyzusi3.nodes import StreamDecoder, AsyncStreamDecoder

class TestMessageDecoderSimple(unittest.TestCase):
    def testDecodeHello(self):
        bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x02\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x46\x61\x68\x72\x70\x75\x6C\x74' + \
            b'\x05\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x32\x2E\x30' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'
        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])        
        self.assertEqual(submessages, [])

        expected_message = messages.HELLO(
            protokollversion=2,
            clienttyp=messages.ClientTyp.FAHRPULT,
            clientname="Fahrpult",
            clientversion="2.0"
        )
        self.assertEqual(basemessage, expected_message)

    def testDecodeAckHello(self):
        bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x09\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x33\x2E\x30\x2E\x31\x2E\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00\xD0\x35\xE4\x40' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'
        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])        
        self.assertEqual(submessages, [])

        expected_message = messages.ACK_HELLO(
            zusiversion = "3.0.1.0",
            verbindungsinfo = "0",
            status = 0,
            startdatum = 41390.5,
            protokollversion = None
        )
        self.assertEqual(basemessage, expected_message)


class TestMessageEncoderSimple(unittest.TestCase):
    def test_encodeObj(self):
        bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x09\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x33\x2E\x30\x2E\x31\x2E\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00\xD0\x35\xE4\x40' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'

        ack_message = messages.ACK_HELLO(
            zusiversion = "3.0.1.0",
            verbindungsinfo = "0",
            status = 0,
            startdatum = 41390.5
        )
    
        encoded_obj = encode_obj(ack_message)
        result = encoded_obj.encode()
        self.assertEqual(result, bytes_written)
    
    def test_encode_complex_msg(self):
        msg = messages.STATUS_INDUSI_EINSTELLUNGEN(indusi_stoerschalter=messages.SCHALTER.EIN)
        msg_bytes = encode_obj(msg).encode()


class TestMessageDecoderEdgeCases(unittest.TestCase):
    def test_ftd169bug(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\xa9\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'
        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 2)
        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[1])        
        # should not raise any issue


class TestMessageEncoderDecoderRoundtrips(unittest.TestCase):
    def test_bool_nodes(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x0a\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x01\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x0b\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x0c\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x03\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x04\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'

        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])
        self.assertEqual(submessages, [])

        expected_basemessage = messages.NEEDED_DATA(
            anzeigen=[messages.FAHRPULT_ANZEIGEN.GESCHWINDIGKEIT],
            bedienung=True,
            programmdaten=[
                messages.PROGRAMMDATEN.ZUGDATEI,
                messages.PROGRAMMDATEN.ZUGNUMMER,
                messages.PROGRAMMDATEN.LADEPAUSE,
                messages.PROGRAMMDATEN.BUCHFAHRPLAN_XML
            ]
        )
        self.assertEqual(basemessage, expected_basemessage)

        encoded_obj = encode_obj(expected_basemessage)
        result = encoded_obj.encode()
        self.assertEqual(result, bytes_written)        

    def test_single_bool_false_msg_reverse(self):
        msg = messages.NEEDED_DATA(
            anzeigen=[
                messages.FAHRPULT_ANZEIGEN.GESCHWINDIGKEIT,
                messages.FAHRPULT_ANZEIGEN.STATUS_SIFA,
                messages.FAHRPULT_ANZEIGEN.STATUS_ZUGBEEINFLUSSUNG
            ],
        )

        encoded_obj = encode_obj(msg)
        encoded_bytes = encoded_obj.encode()

        decoder = StreamDecoder()
        decoded_tree = decoder.decode(encoded_bytes)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])
        self.assertEqual(submessages, [])
        self.assertEqual(basemessage, msg)

    def test_single_bool_true_msg_reverse(self):
        msg = messages.NEEDED_DATA(
            anzeigen=[
                messages.FAHRPULT_ANZEIGEN.GESCHWINDIGKEIT,
                messages.FAHRPULT_ANZEIGEN.STATUS_SIFA,
                messages.FAHRPULT_ANZEIGEN.STATUS_ZUGBEEINFLUSSUNG
            ],
            bedienung=True,
        )

        encoded_obj = encode_obj(msg)
        encoded_bytes = encoded_obj.encode()

        decoder = StreamDecoder()
        decoded_tree = decoder.decode(encoded_bytes)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])
        self.assertEqual(submessages, [])
        self.assertEqual(basemessage, msg)

    def test_msg_nonunique_nodes(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xab\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\xb4\xcc\x0c@' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\xa0@' + \
            b'\x06\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\xff.!G' + \
            b'\x06\x00\x00\x00' + \
            b'\x06\x00' + \
            b'\x00\xa0\x8cF' + \
            b'\x03\x00\x00\x00' + \
            b'\x07\x00' + \
            b'\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\xb4\xcc\x0c@' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\xa0@' + \
            b'\x06\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\xff.!G' + \
            b'\x06\x00\x00\x00' + \
            b'\x06\x00' + \
            b'\x00\xa0\x8cF' + \
            b'\x03\x00\x00\x00' + \
            b'\x07\x00' + \
            b'\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'
        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])

        expected_basemessage = messages.DATA_FTD()
        expected_submessage = messages.STATUS_ZUGFAHRDATEN(
            fahrzeuge = [
                messages.STATUS_ZUGFAHRDATEN_FAHRZEUG(
                    absperrhaehne_hll = messages.ZUGFAHRDATEN_ABSPERRHAEHNE_HLL.STANDARD,
                    bremszylinderdruck=2.1999940872192383,
                    hll_druck=5.0,
                    maximal_moegliche_zugkraft=41262.99609375,
                    maximale_dynamische_bremskraft=18000.0,
                    motordrehzahl_1=0.0,
                    motordrehzahl_2=0.0,
                    zugkraft=0.0),
                messages.STATUS_ZUGFAHRDATEN_FAHRZEUG(
                    absperrhaehne_hll = messages.ZUGFAHRDATEN_ABSPERRHAEHNE_HLL.STANDARD,
                    bremszylinderdruck=2.1999940872192383,
                    hll_druck=5.0,
                    maximal_moegliche_zugkraft=41262.99609375,
                    maximale_dynamische_bremskraft=18000.0,
                    motordrehzahl_1=0.0,
                    motordrehzahl_2=0.0,
                    zugkraft=0.0),
            ]
        )
        self.assertEqual(basemessage, expected_basemessage)
        self.assertEqual(submessages[0], expected_submessage)

        encoded_obj = encode_obj(submessages[0])
        result = encoded_obj.encode()
        self.assertEqual(result, bytes_written)

        encoded_obj = encode_obj(expected_submessage)
        result = encoded_obj.encode()
        self.assertEqual(result, bytes_written)

    def test_lzb_off(self):
        bytes_written = b"\x00\x00\x00\x00" + \
            b"\x02\x00" + \
            b"\x00\x00\x00\x00" + \
            b"\x0a\x01" + \
            b"\x00\x00\x00\x00" + \
            b"\x02\x00" + \
            b"\x00\x00\x00\x00" + \
            b"\x02\x00" + \
            b"\x03\x00\x00\x00" + \
            b"\x07\x00" + \
            b"\x00" + \
            b"\x03\x00\x00\x00" + \
            b"\x08\x00" + \
            b"\x00" + \
            b"\x03\x00\x00\x00" + \
            b"\x09\x00" + \
            b"\x01" + \
            b"\x03\x00\x00\x00" + \
            b"\x0a\x00" + \
            b"\x00" + \
            b"\xFF\xFF\xFF\xFF" + \
            b"\xFF\xFF\xFF\xFF" + \
            b"\xFF\xFF\xFF\xFF" + \
            b"\xFF\xFF\xFF\xFF"

        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)        

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])
        self.assertEqual(submessages, [])

        expected_basemessage = messages.INPUT(
            indusi_hauptschalter=messages.SCHALTER.UNBEKANNT,
            indusi_stoerschalter=messages.SCHALTER.UNBEKANNT,
            indusi_luftabsperrhahn=messages.SCHALTER.UNBEKANNT,
            lzb_stoerschalter=messages.SCHALTER.AUS,
        )
        self.assertEqual(basemessage, expected_basemessage)

        encoded_obj = encode_obj(expected_basemessage)
        result = encoded_obj.encode()
        self.assertEqual(result, bytes_written)

    def test_multisubmessages(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x0b\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x05\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Fbv' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x03\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x03\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x16\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\xa0@' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'v\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00@@' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x14\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Fahr-Bremsschalter' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x0e\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'n\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'u\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00@\xc0' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'N\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b't\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x13\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Richtungsschalter' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'#\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'r\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x0c\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Wagenlicht' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x02\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x11\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Gruppenschalter' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x10\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Schluesselbund' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b's\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x1d\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Federspeicherbremse_Anlegen' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x1c\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Federspeicherbremse_Loesen' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x0b\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Fernlicht' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'x\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff'
        bytes_written += b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x15\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Schlusslicht_Direkt' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x10\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Fenster_Rechts' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'w\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x80?' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x18\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Trittstufenanforderung' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x11\x00\x00\x00' + \
            b'\x01\x00' + \
            b'Scheibenheizung' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x10\x00\x00\x00' + \
            b'\x01\x00' + \
            b'FstBeleuchtung' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x05\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'

        decoder = StreamDecoder()
        decoded_tree = decoder.decode(bytes_written)
        nodes = [node for node in decoded_tree]
        self.assertEqual(len(nodes), 1)        

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])
        pass


class TestAsyncMessageDecoder(unittest. IsolatedAsyncioTestCase):
    async def testDecodeAckHello(self):
        bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x09\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x33\x2E\x30\x2E\x31\x2E\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00\xD0\x35\xE4\x40' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'

        reader = asyncio.StreamReader()
        reader.feed_data(bytes_written)
        reader.feed_eof()

        decoder = AsyncStreamDecoder()
        decoded_tree = decoder.decode(reader)
        nodes = []
        async for node in decoded_tree:
            nodes.append(node)

        self.assertEqual(len(nodes), 1)
        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])        
        self.assertEqual(submessages, [])

        expected_message = messages.ACK_HELLO(
            zusiversion = "3.0.1.0",
            verbindungsinfo = "0",
            status = 0,
            startdatum = 41390.5,
            protokollversion = None
        )
        self.assertEqual(basemessage, expected_message)

    async def test_message_with_empty_strings(self):
        bytes_written = b'\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x0a\x00\x00\x00\x00\x00' + \
            b'\x64\x00\x10\x00\x00\x00\x01\x00\x5a\x65\x69\x74\x2d\x5a\x65\x69' + \
            b'\x74\x2d\x53\x69\x66\x61\x03\x00\x00\x00\x02\x00\x00\x03\x00\x00' + \
            b'\x00\x03\x00\x00\x03\x00\x00\x00\x04\x00\x02\x03\x00\x00\x00\x05' + \
            b'\x00\x02\x03\x00\x00\x00\x06\x00\x02\xff\xff\xff\xff\x00\x00\x00' + \
            b'\x00\x65\x00\x13\x00\x00\x00\x01\x00\x50\x5a\x42\x39\x30\x2f\x49' + \
            b'\x36\x30\x52\x20\x2d\x20\x56\x32\x2e\x30\xff\xff\xff\xff\x00\x00' + \
            b'\x00\x00\x65\x00\x00\x00\x00\x00\x02\x00\x03\x00\x00\x00\x01\x00' + \
            b'\x04\x03\x00\x00\x00\x07\x00\x02\x03\x00\x00\x00\x08\x00\x02\x03' + \
            b'\x00\x00\x00\x0a\x00\x02\x03\x00\x00\x00\x0d\x00\x03\x13\x00\x00' + \
            b'\x00\x0e\x00\x50\x5a\x42\x39\x30\x2f\x49\x36\x30\x52\x20\x2d\x20' + \
            b'\x56\x32\x2e\x30\xff\xff\xff\xff\x00\x00\x00\x00\x03\x00\x04\x00' + \
            b'\x00\x00\x02\x00\x05\x00\x04\x00\x00\x00\x03\x00\x00\x00\x03\x00' + \
            b'\x00\x00\x05\x00\x00\x03\x00\x00\x00\x09\x00\x00\x03\x00\x00\x00' + \
            b'\x2f\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x65' + \
            b'\x00\x00\x00\x00\x00\x02\x00\x02\x00\x00\x00\x02\x00\x02\x00\x00' + \
            b'\x00\x03\x00\x03\x00\x00\x00\x0b\x00\x00\x03\x00\x00\x00\x0f\x00' + \
            b'\x00\x00\x00\x00\x00\x05\x00\x04\x00\x00\x00\x01\x00\xa0\x00\x04' + \
            b'\x00\x00\x00\x02\x00\x09\x00\x03\x00\x00\x00\x05\x00\x04\xff\xff' + \
            b'\xff\xff\x00\x00\x00\x00\x06\x00\x04\x00\x00\x00\x01\x00\xa2\x00' + \
            b'\x04\x00\x00\x00\x02\x00\x08\x00\x03\x00\x00\x00\x05\x00\x04\x03' + \
            b'\x00\x00\x00\x06\x00\x06\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00' + \
            b'\x00\x00\x03\x00\x03\x00\x00\x00\x06\x00\x00\x03\x00\x00\x00\x07' + \
            b'\x00\x00\x03\x00\x00\x00\x08\x00\x01\x03\x00\x00\x00\x0a\x00\x00' + \
            b'\x03\x00\x00\x00\x0b\x00\x00\x03\x00\x00\x00\x30\x00\x02\x03\x00' + \
            b'\x00\x00\x31\x00\x03\x03\x00\x00\x00\x32\x00\x00\x03\x00\x00\x00' + \
            b'\x33\x00\x00\x03\x00\x00\x00\x34\x00\x00\xff\xff\xff\xff\xff\xff' + \
            b'\xff\xff\x00\x00\x00\x00\x65\x00\x00\x00\x00\x00\x03\x00\x03\x00' + \
            b'\x00\x00\x0c\x00\x02\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff\xff\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x0a' + \
            b'\x00\x06\x00\x00\x00\x01\x00\x00\x00\x00\x00\xff\xff\xff\xff\xff' + \
            b'\xff\xff\xff'
        reader = asyncio.StreamReader()
        reader.feed_data(bytes_written)
        reader.feed_eof()

        decoder = AsyncStreamDecoder()
        decoded_tree = decoder.decode(reader)
        nodes = []
        async for node in decoded_tree:
            nodes.append(node)

        self.assertEqual(len(nodes), 2)

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[0])

        messagedecoder = MessageDecoder()
        basemessage, submessages = messagedecoder.parse(nodes[1])
        


class BaseException(Exception):
    """Base class to all exceptions produced in the lib"""
    pass


class MissingContentTypeError(BaseException):
    """Raised when content for a node is set but the type is missing"""
    pass


class EncodingValueError(BaseException):
    """Raised when content cannot be encoded to required bytesize"""
    pass


class MissingLowLevelParameterError(BaseException):
    """Raised when given class of object has no matching encoding rule"""
    pass


class MissingBytesDecodeError(BaseException):
    """Raised when binary state does not match expected content"""
    pass


class DecodeValueError(BaseException):
    """Raised when value cannot be decoded to requested content type"""
    pass
